import numpy as np
import copy
from skretrieval.core import OpticalGeometry
from sasktran import Geometry, StokesVector
from abc import ABC, abstractmethod
from typing import List
import xarray as xr


class OpticalComponent(ABC):
    """
    Abstract optical element that takes a radiance and tranforms it based on its Mueller matrix. Designed to
    be placed (or stacked) in front of `Sensor` objects to transform radiance before reaching the detector.
    """

    @abstractmethod
    def matrix(self, wavelength) -> np.array:
        """
        Returns
        -------
        np.ndarray
            4 x 4 mueller matrix
        """
        raise NotImplemented()

    def __add__(self, other):
        if type(other) is CompositeComponent:
            return CompositeComponent([self] + other.components)
        else:
            return CompositeComponent([self, other])

    def model_radiance(self, optical_geometry: OpticalGeometry,
                       model_wavel_nm: np.array = None,
                       model_geometry: Geometry = None,
                       radiance: np.array = None,
                       wf: np.array = None):

        if type(radiance) is np.ndarray:
            # make a deepcopy since array elements are objects
            new_rad = copy.deepcopy(radiance)
            transmission = np.ones(len(model_wavel_nm))
            for widx, rad in enumerate(new_rad):
                mueller = self.matrix(model_wavel_nm[widx])
                transmission[widx] = mueller[0, 0]
                for lidx, r in enumerate(rad):
                    new_basis = self.stokes_basis_from_geometry(optical_geometry, r)
                    r.to_new_basis(new_basis)
                    stokes = mueller @ r._stokes
                    new_rad[widx, lidx] = StokesVector(stokes, new_basis)

        elif type(radiance) is xr.Dataset:
            # new_rad = np.ones((len(radiance.wavelength), len(radiance.los)), dtype=StokesVector)
            if wf is None:
                new_rad = radiance.copy(deep=True).transpose('wavelength', 'los', 'xyz')
            else:
                new_rad = radiance.copy(deep=True).transpose('wavelength', 'los', 'xyz', 'perturbation')

            I = radiance.I.values
            Q = radiance.Q.values
            U = radiance.U.values
            V = radiance.V.values
            prop_dir = radiance.propagation_direction
            theta_dir = radiance.theta_direction
            phi_dir = radiance.phi_direction

            transmission = xr.ones_like(radiance.wavelength)
            for widx, wavel in enumerate(radiance.wavelength.values):
                print('wavelength', wavel)
                mueller = self.matrix(model_wavel_nm[widx])
                transmission.values[widx] = mueller[0, 0]
                for lidx, los in enumerate(radiance.los.values):
                    r = StokesVector([float(I[widx, lidx]), float(Q[widx, lidx]), float(U[widx, lidx]), float(V[widx, lidx])],
                                     np.array([prop_dir[lidx], theta_dir[lidx], phi_dir[lidx]]))
                    # if widx == 0:
                    new_basis = self.stokes_basis_from_geometry(optical_geometry, r)
                    r.to_new_basis(new_basis)
                    stokes = mueller @ r._stokes
                    sv = StokesVector(stokes, new_basis)
                    new_rad.I.values[widx, lidx] = sv.I
                    new_rad.Q.values[widx, lidx] = sv.Q
                    new_rad.U.values[widx, lidx] = sv.U
                    new_rad.V.values[widx, lidx] = sv.V
                    new_rad.propagation_direction[lidx] = sv.propagation_direction
                    new_rad.theta_direction[lidx] = sv.theta_direction
                    new_rad.phi_direction[lidx] = sv.phi_direction

        if wf is None:
            return new_rad
        else:
            return new_rad, self.update_weighting_function(wf, radiance, new_rad, transmission)

    @staticmethod
    def update_weighting_function(wf, old_radiance, new_radiance, transmission=1.0):
        # TODO: I think the wf needs to be scaled by the transmission?
        if type(wf) is xr.DataArray:
            return wf * new_radiance.I / old_radiance.I / transmission
        elif type(wf) is xr.Dataset:
            return wf * new_radiance.I / old_radiance.I / transmission
        else:
            return wf * np.array([(n.I / r.I) for n, r
                                  in zip(new_radiance.flatten(),
                                         old_radiance.flatten())]).reshape(old_radiance.shape)[:, :, np.newaxis] / transmission

    @staticmethod
    def stokes_basis_from_geometry(optical_geometry: OpticalGeometry, stokes_vector: StokesVector) -> np.ndarray:
        """
        Create a basis with local_up and horizontal axis of optical axis projected onto the plane defined by
        the stokes vector propagation direction.

        Parameters
        ----------
        optical_geometry : OpticalGeometry
        stokes_vector : StokesVector
            Axis along which optical_geometry will be projected.

        Returns
        -------
        np.ndarray

        """
        p = stokes_vector.propagation_direction
        up = optical_geometry.local_up - np.dot(optical_geometry.local_up, p) * p
        up /= np.linalg.norm(up)

        x = np.cross(p, up)
        x /= np.linalg.norm(x)

        basis = np.array([p, up, x])
        return basis


class CompositeComponent(OpticalComponent):

    def __init__(self, components: List[OpticalComponent]):
        """
        Used to compose multiple `OpticalComponents` into a single element for faster processing. This way only one
        Mueller matrix is needed and the basis is only transformed once.

        Parameters
        ----------

        components : List[OpticalComponent]
            List of `OpticalComponent`s in the order they are arranged in the system

        Examples
        --------

        Create a circular retarder::

            rotator = CompositeComponent([HalfWavePlate(45), HalfWavePlate(0)])

        `CompositeComponent`s can also be constructed by adding components::

            rotator = HalfWavePlate(45) + HalfWavePlate(0)

        """
        self._components = components

    def __add__(self, other):

        if type(other) is CompositeComponent:
            return CompositeComponent(self._components + other.components)
        else:
            return CompositeComponent(self._components + [other])

    @property
    def components(self):
        return self._components

    def matrix(self, wavelength: float) -> np.array:

        m = np.eye(4)
        for component in self._components[::-1]:
            m = m @ component.matrix(wavelength)

        return m


class OpticalComponentFromG(OpticalComponent):

    def __init__(self, wavelength: np.array, g1: np.array = None, g2: np.array = None, g3: np.array = None):

        self._wavelength = wavelength
        self._g1 = g1
        self._g2 = g2
        self._g3 = g3

    def _interp_g(self, g, w, wavelength):
        if g is None:
            return 0
        else:
            return np.interp(wavelength, w, g, left=g[0], right=g[-1])

    def g1(self, wavelength: float):
        return self._interp_g(self._g1, self._wavelength, wavelength)

    def g2(self, wavelength: float):
        return self._interp_g(self._g1, self._wavelength, wavelength)

    def g3(self, wavelength: float):
        return self._interp_g(self._g1, self._wavelength, wavelength)

    def matrix(self, wavelength: float) -> np.array:

        matrix = np.zeros((4, 4))
        matrix[0, 0] = 1
        matrix[0, 1] = self.g1(wavelength)
        matrix[0, 2] = self.g2(wavelength)
        matrix[0, 3] = self.g3(wavelength)
        return matrix


class ClearGlass(OpticalComponent):
    """
    Clear glass component that does nothing but transform the radiance into a new base
    """
    def matrix(self, wavelength) -> np.array:
        return np.eye(4, dtype=np.float)
