from skcomponents.processing.processor import SignalProcessor
import numpy as np
from typing import Union


class PhotonIntegration(SignalProcessor):

    def __init__(self, aperture_area_cm2: float,
                 num_rows: int, num_cols: int,
                 horizontal_fov: float, vertical_fov: float,
                 integration_time: float = 1,
                 wavelength_area: Union[float, np.ndarray] = 1):
        """
        Integrate an radiance field over the pixel areas and wavelength in a 2D detector.
        Radiance is assumed to be unchanging over the pixel area.

        Parameters
        ----------
        aperture_area_cm2 : float
            Area of the light limiting aperture
        num_rows : float
            Number of rows in the image
        num_cols : float
            Number of columns in the image
        horizontal_fov : float
            Horizontal field of view of the full image
        vertical_fov : float
            Vertical field of view of the full image
        integration_time : float
            Length of exposure time in seconds
        wavelength_area : float or np.ndarray
            Integrated spectral line shape area. If an array is provided, then it is assumed to be the same length as
            number of wavelengths in the radiance calculation
        """

        self.aperture_area = aperture_area_cm2
        self.num_pixels = num_cols * num_rows
        self.solid_angle = (2 * np.sin(vertical_fov * np.pi / 180 / 2)) * (horizontal_fov * np.pi / 180)  # detector solid angle
        self.pixel_solid_angle = self.solid_angle / self.num_pixels
        self.projected_pixel_area = self.aperture_area / self.num_pixels
        self.integration_time = integration_time
        self.wavelength_area = wavelength_area

    @property
    def scale_factor(self):
        return self.projected_pixel_area * self.solid_angle * self.integration_time * self.wavelength_area

    def process_signal(self, radiance: np.ndarray) -> np.ndarray:
        """

        Parameters
        ----------
        radiance : np.ndarray
            irradiance is assumed to be a grid that spans the size of the sensor. This can be achieved by using a sensor
            with an unnormalized spectral PSF, and a normalized spatial PSF to integrate front-end radiance at the front
            end aperture.

        Returns
        -------
        np.ndarray : irradiance integrated over pixel area
        """

        if radiance.shape[-1] == self.scale_factor.shape[0]:
            return radiance * self.scale_factor
        elif radiance.shape[0] == self.scale_factor.shape[0]:
            return radiance * self.scale_factor[:, np.newaxis, np.newaxis]
        else:
            raise ValueError('Cannot broadcast radiance and scale_factor, '
                             'wavelength should be the first or last dimension')


class PhotonsToElectrons(SignalProcessor):

    def __init__(self, conversion_efficiency: float = 1.0, well_capacity: int = None):

        self.conversion_efficiency = conversion_efficiency
        self.well_capacity = well_capacity

    def process_signal(self, signal: np.ndarray) -> np.ndarray:

        if self.well_capacity is not None:
            signal[signal > self.well_capacity] = self.well_capacity

        return signal * self.conversion_efficiency


class ElectronsToVoltage(SignalProcessor):

    def __init__(self, capacitance: float, gain: float):

        self.capacitance = capacitance
        self.gain = gain

    def process_signal(self, signal: np.ndarray) -> np.ndarray:

        return signal / self.capacitance * self.gain
