import pytest
import numpy as np
import xarray as xr
from sasktran import StokesVector, Geometry, LineOfSight
from skretrieval.core import OpticalGeometry


@pytest.fixture
def radiance():
    return np.array([[StokesVector(np.array([1, 1, 0, 0], dtype=np.float), basis=np.eye(3)),
                      StokesVector(np.array([1, -1, 0, 0], dtype=np.float), basis=np.eye(3)),
                      StokesVector(np.array([1, 0, 1, 0], dtype=np.float), basis=np.eye(3)),
                      StokesVector(np.array([1, 0, -1, 0], dtype=np.float), basis=np.eye(3))]])


@pytest.fixture
def radiance_dataset(radiance, model_wavelength):

    ds = xr.Dataset({'I': (['wavelength', 'los'], np.array([[r.I for r in rad] for rad in radiance])),
                     'Q': (['wavelength', 'los'], np.array([[float(r.Q) for r in rad] for rad in radiance])),
                     'U': (['wavelength', 'los'], np.array([[float(r.U) for r in rad] for rad in radiance])),
                     'V': (['wavelength', 'los'], np.array([[float(r.V) for r in rad] for rad in radiance])),
                     'propagation_direction': (['los', 'xyz'], np.array([r.propagation_direction for r in radiance[0]])),
                     'theta_direction': (['los', 'xyz'], np.array([r.theta_direction for r in radiance[0]])),
                     'phi_direction': (['los', 'xyz'], np.array([r.phi_direction for r in radiance[0]]))},
                    coords={'wavelength': model_wavelength,
                            'los': np.arange(0, len(radiance[0])),
                            'xyz': ['x', 'y', 'z']})
    return ds

@pytest.fixture
def model_geometry():
    geom = Geometry()
    geom.lines_of_sight = [LineOfSight(mjd=53000.0, observer=np.array([1, 0, 0]), look_vector=np.array([0, 1, 0]))]
    return geom


@pytest.fixture
def optical_geometry():
    return OpticalGeometry(observer=np.array([1, 0, 0]), look_vector=np.array([0, 1, 0]),
                           local_up=np.array([0, 0, 1]), mjd=53000.0)


@pytest.fixture
def model_wavelength():
    return [500.0]
